theory Semantics imports Main begin

section {* The Language *}

subsection {* Variables and Values *}

types vname = string -- "names for variables"

datatype val
  = Bool bool      -- "Boolean value"
  | Intg int       -- "integer value" 

syntax
  true  :: val
  false :: val
translations
  "true"  == "Bool True"
  "false" == "Bool False"



subsection {* Expressions and Commands*}

datatype bop = Eq | And | Less | Add | Sub     -- "names of binary operations"

datatype expr
  = Val val                                          -- "value"
  | Var vname                                        -- "local variable"
  | BinOp expr bop expr    ("_ \<guillemotleft>_\<guillemotright> _" [80,0,81] 80)  -- "binary operation"


text {* Note: we assume that only type correct expressions are regarded
  as later proofs fail if expressions evaluate to None due to type errors.
  However there is [yet] no typing system *}

fun binop :: "bop \<Rightarrow> val \<Rightarrow> val \<Rightarrow> val option"
where "binop Eq v\<^isub>1 v\<^isub>2               = Some(Bool(v\<^isub>1 = v\<^isub>2))"
  | "binop And (Bool b\<^isub>1) (Bool b\<^isub>2)  = Some(Bool(b\<^isub>1 \<and> b\<^isub>2))"
  | "binop Less (Intg i\<^isub>1) (Intg i\<^isub>2) = Some(Bool(i\<^isub>1 < i\<^isub>2))"
  | "binop Add (Intg i\<^isub>1) (Intg i\<^isub>2)  = Some(Intg(i\<^isub>1 + i\<^isub>2))"
  | "binop Sub (Intg i\<^isub>1) (Intg i\<^isub>2)  = Some(Intg(i\<^isub>1 - i\<^isub>2))"

  | "binop bop v\<^isub>1 v\<^isub>2                = Some(Intg(0))"


datatype com
  = Skip
  | LAss vname expr        ("_:=_" [70,70] 70)  -- "local assignment"
  | Seq com com            ("_;;/ _" [61,60] 60)
  | Cond expr com com      ("if '(_') _/ else _" [80,79,79] 70)
  | While expr com         ("while '(_') _" [80,79] 70)


fun fv :: "expr \<Rightarrow> vname set" --"free variables in an expression"
where
  FVc: "fv (Val V) = {}"
  | FVv: "fv (Var V) = {V}"
  | FVe: "fv (e1 \<guillemotleft>bop\<guillemotright> e2) = fv e1 \<union> fv e2"


subsection {* State *}

types
  state = "vname \<rightharpoonup> val"


text {* @{text interpret} silently assumes type correct expressions, 
  i.e. no expression evaluates to None *}

fun "interpret" :: "expr \<Rightarrow> state \<Rightarrow> val option" ("\<lbrakk>_\<rbrakk>_")
where Val: "\<lbrakk>Val v\<rbrakk> s = Some v"
  | Var: "\<lbrakk>Var V\<rbrakk> s = s V"
  | BinOp: "\<lbrakk>e\<^isub>1\<guillemotleft>bop\<guillemotright>e\<^isub>2\<rbrakk> s = (case \<lbrakk>e\<^isub>1\<rbrakk> s of None \<Rightarrow> None
    | Some v\<^isub>1 \<Rightarrow> (case \<lbrakk>e\<^isub>2\<rbrakk> s of None \<Rightarrow> None
                           | Some v\<^isub>2 \<Rightarrow> binop bop v\<^isub>1 v\<^isub>2))"

subsection {* Small Step Semantics *}

inductive red :: "com * state \<Rightarrow> com * state \<Rightarrow> bool"
and red' :: "com \<Rightarrow> state \<Rightarrow> com \<Rightarrow> state \<Rightarrow> bool"
   ("((1\<langle>_,/_\<rangle>) \<rightarrow>/ (1\<langle>_,/_\<rangle>))" [0,0,0,0] 81)
where
  "\<langle>c1,s1\<rangle> \<rightarrow> \<langle>c2,s2\<rangle> == red (c1,s1) (c2,s2)" |
  RedLAss:
  "\<langle>V:=e,s\<rangle> \<rightarrow> \<langle>Skip,s(V:=(\<lbrakk>e\<rbrakk> s))\<rangle>"

  | SeqRed:
  "\<langle>c\<^isub>1,s\<rangle> \<rightarrow> \<langle>c\<^isub>1',s'\<rangle> \<Longrightarrow> \<langle>c\<^isub>1;;c\<^isub>2,s\<rangle> \<rightarrow> \<langle>c\<^isub>1';;c\<^isub>2,s'\<rangle>"

  | RedSeq:
  "\<langle>Skip;;c\<^isub>2,s\<rangle> \<rightarrow> \<langle>c\<^isub>2,s\<rangle>"

  | RedCondTrue:
  "\<lbrakk>b\<rbrakk> s = Some true \<Longrightarrow> \<langle>if (b) c\<^isub>1 else c\<^isub>2,s\<rangle> \<rightarrow> \<langle>c\<^isub>1,s\<rangle>"

  | RedCondFalse:
  "\<lbrakk>b\<rbrakk> s = Some false \<Longrightarrow> \<langle>if (b) c\<^isub>1 else c\<^isub>2,s\<rangle> \<rightarrow> \<langle>c\<^isub>2,s\<rangle>"

  | RedWhileTrue:
  "\<lbrakk>b\<rbrakk> s = Some true \<Longrightarrow> \<langle>while (b) c,s\<rangle> \<rightarrow> \<langle>c;;while (b) c,s\<rangle>"

  | RedWhileFalse:
  "\<lbrakk>b\<rbrakk> s = Some false \<Longrightarrow> \<langle>while (b) c,s\<rangle> \<rightarrow> \<langle>Skip,s\<rangle>"

lemmas red_induct = red.induct[split_format (complete)]

abbreviation reds ::"com \<Rightarrow> state \<Rightarrow> com \<Rightarrow> state \<Rightarrow> bool" 
   ("((1\<langle>_,/_\<rangle>) \<rightarrow>*/ (1\<langle>_,/_\<rangle>))" [0,0,0,0] 81) where
  "\<langle>c,s\<rangle> \<rightarrow>* \<langle>c',s'\<rangle> ==  red\<^sup>*\<^sup>* (c,s) (c',s')"

lemma Skip_reds:
  "\<langle>Skip,s\<rangle> \<rightarrow>* \<langle>c',s'\<rangle> \<Longrightarrow> s = s' \<and> c' = Skip"
by(blast elim:converse_rtranclpE red.cases)

lemma LAss_reds:
  "\<langle>V:=e,s\<rangle> \<rightarrow>* \<langle>Skip,s'\<rangle> \<Longrightarrow> s' = s(V:=\<lbrakk>e\<rbrakk> s)"
proof(induct c=="V:=e" s rule:converse_rtranclp_induct2)
  case refl thus ?case by simp
next
  case (step c s c'' s'')
  hence "c'' = Skip" and "s'' = s(V:=(\<lbrakk>e\<rbrakk> s))" by(auto elim:red.cases)
  with `\<langle>c'',s''\<rangle> \<rightarrow>* \<langle>Skip,s'\<rangle>` show ?case by(auto dest:Skip_reds)
qed

lemma Seq2_reds:
  "\<langle>Skip;;c\<^isub>2,s\<rangle> \<rightarrow>* \<langle>Skip,s'\<rangle> \<Longrightarrow> \<langle>c\<^isub>2,s\<rangle> \<rightarrow>* \<langle>Skip,s'\<rangle>"
by(induct c=="Skip;;c\<^isub>2" s rule:converse_rtranclp_induct2)(auto elim:red.cases)

lemma Seq_reds:
  assumes "\<langle>c\<^isub>1;;c\<^isub>2,s\<rangle> \<rightarrow>* \<langle>Skip,s'\<rangle>"
  obtains s'' where "\<langle>c\<^isub>1,s\<rangle> \<rightarrow>* \<langle>Skip,s''\<rangle>" and "\<langle>c\<^isub>2,s''\<rangle> \<rightarrow>* \<langle>Skip,s'\<rangle>"
proof -
  have "\<exists>s''. \<langle>c\<^isub>1,s\<rangle> \<rightarrow>* \<langle>Skip,s''\<rangle> \<and> \<langle>c\<^isub>2,s''\<rangle> \<rightarrow>* \<langle>Skip,s'\<rangle>"
  proof -
    { fix c c'
      assume "\<langle>c,s\<rangle> \<rightarrow>* \<langle>c',s'\<rangle>" and "c = c\<^isub>1;;c\<^isub>2" and "c' = Skip"
      hence "\<exists>s''. \<langle>c\<^isub>1,s\<rangle> \<rightarrow>* \<langle>Skip,s''\<rangle> \<and> \<langle>c\<^isub>2,s''\<rangle> \<rightarrow>* \<langle>Skip,s'\<rangle>"
      proof(induct arbitrary:c\<^isub>1 rule:converse_rtranclp_induct2)
	case refl thus ?case by simp
      next
	case (step c s c'' s'')
	note IH = `\<And>c\<^isub>1. \<lbrakk>c'' = c\<^isub>1;;c\<^isub>2; c' = Skip\<rbrakk> 
	  \<Longrightarrow> \<exists>sx. \<langle>c\<^isub>1,s''\<rangle> \<rightarrow>* \<langle>Skip,sx\<rangle> \<and> \<langle>c\<^isub>2,sx\<rangle> \<rightarrow>* \<langle>Skip,s'\<rangle>`
	from step
	have "\<langle>c\<^isub>1;;c\<^isub>2,s\<rangle> \<rightarrow> \<langle>c'',s''\<rangle>" by simp
	hence "(c\<^isub>1 = Skip \<and> c'' = c\<^isub>2 \<and> s = s'') \<or> 
	  (\<exists>c\<^isub>1'. \<langle>c\<^isub>1,s\<rangle> \<rightarrow> \<langle>c\<^isub>1',s''\<rangle> \<and> c'' = c\<^isub>1';;c\<^isub>2)"
	  by(auto elim:red.cases)
	thus ?case
	proof
	  assume "c\<^isub>1 = Skip \<and> c'' = c\<^isub>2 \<and> s = s''"
	  with `\<langle>c'',s''\<rangle> \<rightarrow>* \<langle>c',s'\<rangle>` `c' = Skip`
	  show ?thesis by auto
	next
	  assume "\<exists>c\<^isub>1'. \<langle>c\<^isub>1,s\<rangle> \<rightarrow> \<langle>c\<^isub>1',s''\<rangle> \<and> c'' = c\<^isub>1';;c\<^isub>2"
	  then obtain c\<^isub>1' where "\<langle>c\<^isub>1,s\<rangle> \<rightarrow> \<langle>c\<^isub>1',s''\<rangle>" and "c'' = c\<^isub>1';;c\<^isub>2" by blast
	  from IH[OF `c'' = c\<^isub>1';;c\<^isub>2` `c' = Skip`]
	  obtain sx where "\<langle>c\<^isub>1',s''\<rangle> \<rightarrow>* \<langle>Skip,sx\<rangle>" and "\<langle>c\<^isub>2,sx\<rangle> \<rightarrow>* \<langle>Skip,s'\<rangle>"
	    by blast
	  from `\<langle>c\<^isub>1,s\<rangle> \<rightarrow> \<langle>c\<^isub>1',s''\<rangle>` `\<langle>c\<^isub>1',s''\<rangle> \<rightarrow>* \<langle>Skip,sx\<rangle>`
	  have "\<langle>c\<^isub>1,s\<rangle> \<rightarrow>* \<langle>Skip,sx\<rangle>" by(auto intro:converse_rtranclp_into_rtranclp)
	  with `\<langle>c\<^isub>2,sx\<rangle> \<rightarrow>* \<langle>Skip,s'\<rangle>` show ?thesis by auto
	qed
      qed }
    with `\<langle>c\<^isub>1;;c\<^isub>2,s\<rangle> \<rightarrow>* \<langle>Skip,s'\<rangle>` show ?thesis by simp
  qed
  with that show ?thesis by blast
qed


lemma Cond_True_or_False:
  "\<langle>if (b) c\<^isub>1 else c\<^isub>2,s\<rangle> \<rightarrow>* \<langle>Skip,s'\<rangle> \<Longrightarrow> \<lbrakk>b\<rbrakk> s = Some true \<or> \<lbrakk>b\<rbrakk> s = Some false"
by(induct c=="if (b) c\<^isub>1 else c\<^isub>2" s rule:converse_rtranclp_induct2)(auto elim:red.cases)

lemma CondTrue_reds:
  "\<langle>if (b) c\<^isub>1 else c\<^isub>2,s\<rangle> \<rightarrow>* \<langle>Skip,s'\<rangle> \<Longrightarrow> \<lbrakk>b\<rbrakk> s = Some true \<Longrightarrow> \<langle>c\<^isub>1,s\<rangle> \<rightarrow>* \<langle>Skip,s'\<rangle>"
by(induct c=="if (b) c\<^isub>1 else c\<^isub>2" s rule:converse_rtranclp_induct2)(auto elim:red.cases)

lemma CondFalse_reds:
  "\<langle>if (b) c\<^isub>1 else c\<^isub>2,s\<rangle> \<rightarrow>* \<langle>Skip,s'\<rangle> \<Longrightarrow> \<lbrakk>b\<rbrakk> s = Some false \<Longrightarrow> \<langle>c\<^isub>2,s\<rangle> \<rightarrow>* \<langle>Skip,s'\<rangle>"
by(induct c=="if (b) c\<^isub>1 else c\<^isub>2" s rule:converse_rtranclp_induct2)(auto elim:red.cases)

lemma WhileFalse_reds: 
  "\<langle>while (b) cx,s\<rangle> \<rightarrow>* \<langle>Skip,s'\<rangle> \<Longrightarrow> \<lbrakk>b\<rbrakk> s = Some false \<Longrightarrow> s = s'"
proof(induct c=="while (b) cx" s rule:converse_rtranclp_induct2)
  case refl thus ?case by simp
next
  case step thus ?case by(auto elim:red.cases dest: Skip_reds)
qed

lemma WhileTrue_reds: 
  "\<langle>while (b) cx,s\<rangle> \<rightarrow>* \<langle>Skip,s'\<rangle> \<Longrightarrow> \<lbrakk>b\<rbrakk> s = Some true
  \<Longrightarrow> \<exists>sx. \<langle>cx,s\<rangle> \<rightarrow>* \<langle>Skip,sx\<rangle> \<and> \<langle>while (b) cx,sx\<rangle> \<rightarrow>* \<langle>Skip,s'\<rangle>"
proof(induct c=="while (b) cx" s rule:converse_rtranclp_induct2)
  case refl thus ?case by simp
next
  case (step c s c'' s'')
  hence "c'' = cx;;while (b) cx \<and> s'' = s" by(auto elim:red.cases)
  with `\<langle>c'',s''\<rangle> \<rightarrow>* \<langle>Skip,s'\<rangle>` show ?case by(auto dest:Seq_reds)
qed

lemma While_True_or_False:
  "\<langle>while (b) com,s\<rangle> \<rightarrow>* \<langle>Skip,s'\<rangle> \<Longrightarrow> \<lbrakk>b\<rbrakk> s = Some true \<or> \<lbrakk>b\<rbrakk> s = Some false"
by(induct c=="while (b) com" s rule:converse_rtranclp_induct2)(auto elim:red.cases)


inductive red_n :: "com \<Rightarrow> state \<Rightarrow> nat \<Rightarrow> com \<Rightarrow> state \<Rightarrow> bool"
   ("((1\<langle>_,/_\<rangle>) \<rightarrow>\<^bsup>_\<^esup> (1\<langle>_,/_\<rangle>))" [0,0,0,0,0] 81)
where red_n_Base: "\<langle>c,s\<rangle> \<rightarrow>\<^bsup>0\<^esup> \<langle>c,s\<rangle>"
     | red_n_Rec: "\<lbrakk>\<langle>c,s\<rangle> \<rightarrow> \<langle>c'',s''\<rangle>; \<langle>c'',s''\<rangle> \<rightarrow>\<^bsup>n\<^esup> \<langle>c',s'\<rangle>\<rbrakk> \<Longrightarrow> \<langle>c,s\<rangle> \<rightarrow>\<^bsup>Suc n\<^esup> \<langle>c',s'\<rangle>"


lemma Seq_red_nE: assumes "\<langle>c\<^isub>1;;c\<^isub>2,s\<rangle> \<rightarrow>\<^bsup>n\<^esup> \<langle>Skip,s'\<rangle>"
  obtains i j s'' where "\<langle>c\<^isub>1,s\<rangle> \<rightarrow>\<^bsup>i\<^esup> \<langle>Skip,s''\<rangle>" and "\<langle>c\<^isub>2,s''\<rangle> \<rightarrow>\<^bsup>j\<^esup> \<langle>Skip,s'\<rangle>"
  and "n = i + j + 1"
proof -
  from `\<langle>c\<^isub>1;;c\<^isub>2,s\<rangle> \<rightarrow>\<^bsup>n\<^esup> \<langle>Skip,s'\<rangle>` 
  have "\<exists>i j s''. \<langle>c\<^isub>1,s\<rangle> \<rightarrow>\<^bsup>i\<^esup> \<langle>Skip,s''\<rangle> \<and> \<langle>c\<^isub>2,s''\<rangle> \<rightarrow>\<^bsup>j\<^esup> \<langle>Skip,s'\<rangle> \<and> n = i + j + 1"
  proof(induct c\<equiv>"c\<^isub>1;;c\<^isub>2" s n c'\<equiv>"Skip" s' arbitrary:c\<^isub>1 rule:red_n.induct)
    case (red_n_Base c s)
    from `c = c\<^isub>1;;c\<^isub>2` `c = Skip` have False by simp
    thus ?case by simp
  next
    case (red_n_Rec c s c'' s'' n c' s')
    note IH = `\<And>c\<^isub>1. \<lbrakk>c'' = c\<^isub>1;;c\<^isub>2; c' = Skip\<rbrakk>
      \<Longrightarrow> \<exists>i j sx. \<langle>c\<^isub>1,s''\<rangle> \<rightarrow>\<^bsup>i\<^esup> \<langle>Skip,sx\<rangle> \<and> \<langle>c\<^isub>2,sx\<rangle> \<rightarrow>\<^bsup>j\<^esup> \<langle>Skip,s'\<rangle> \<and> n = i + j + 1`
    from `\<langle>c,s\<rangle> \<rightarrow> \<langle>c'',s''\<rangle>` `c = c\<^isub>1;;c\<^isub>2`
    have "(c\<^isub>1 = Skip \<and> c'' = c\<^isub>2 \<and> s = s'') \<or> 
      (\<exists>c\<^isub>1'. c'' = c\<^isub>1';;c\<^isub>2 \<and> \<langle>c\<^isub>1,s\<rangle> \<rightarrow> \<langle>c\<^isub>1',s''\<rangle>)"
      by(induct rule:red_induct) auto
    thus ?case
    proof
      assume "c\<^isub>1 = Skip \<and> c'' = c\<^isub>2 \<and> s = s''"
      hence "c\<^isub>1 = Skip" and "c'' = c\<^isub>2" and "s = s''" by simp_all
      from `c\<^isub>1 = Skip` have "\<langle>c\<^isub>1,s\<rangle> \<rightarrow>\<^bsup>0\<^esup> \<langle>Skip,s\<rangle>" by(fastsimp intro:red_n_Base)
      with `\<langle>c'',s''\<rangle> \<rightarrow>\<^bsup>n\<^esup> \<langle>c',s'\<rangle>` `c'' = c\<^isub>2` `c' = Skip` `s = s''`
      show ?thesis by(rule_tac x="0" in exI) auto
    next
      assume "\<exists>c\<^isub>1'. c'' = c\<^isub>1';;c\<^isub>2 \<and> \<langle>c\<^isub>1,s\<rangle> \<rightarrow> \<langle>c\<^isub>1',s''\<rangle>"
      then obtain c\<^isub>1' where "c'' = c\<^isub>1';;c\<^isub>2" and "\<langle>c\<^isub>1,s\<rangle> \<rightarrow> \<langle>c\<^isub>1',s''\<rangle>" by blast
      from IH[OF `c'' = c\<^isub>1';;c\<^isub>2` `c' = Skip`] obtain i j sx
	where "\<langle>c\<^isub>1',s''\<rangle> \<rightarrow>\<^bsup>i\<^esup> \<langle>Skip,sx\<rangle>" and "\<langle>c\<^isub>2,sx\<rangle> \<rightarrow>\<^bsup>j\<^esup> \<langle>Skip,s'\<rangle>"
	and "n = i + j + 1" by blast
      from `\<langle>c\<^isub>1,s\<rangle> \<rightarrow> \<langle>c\<^isub>1',s''\<rangle>` `\<langle>c\<^isub>1',s''\<rangle> \<rightarrow>\<^bsup>i\<^esup> \<langle>Skip,sx\<rangle>`
      have "\<langle>c\<^isub>1,s\<rangle> \<rightarrow>\<^bsup>Suc i\<^esup> \<langle>Skip,sx\<rangle>" by(rule red_n.red_n_Rec)
      with `\<langle>c\<^isub>2,sx\<rangle> \<rightarrow>\<^bsup>j\<^esup> \<langle>Skip,s'\<rangle>` `n = i + j + 1` show ?thesis
	by(rule_tac x="Suc i" in exI) auto
    qed
  qed
  with that show ?thesis by blast
qed


lemma while_red_nE:
  "\<langle>while (b) cx,s\<rangle> \<rightarrow>\<^bsup>n\<^esup> \<langle>Skip,s'\<rangle> 
  \<Longrightarrow> (\<lbrakk>b\<rbrakk> s = Some false \<and> s = s' \<and> n = 1) \<or>
     (\<exists>i j s''. \<lbrakk>b\<rbrakk> s = Some true \<and> \<langle>cx,s\<rangle> \<rightarrow>\<^bsup>i\<^esup> \<langle>Skip,s''\<rangle> \<and> 
                \<langle>while (b) cx,s''\<rangle> \<rightarrow>\<^bsup>j\<^esup> \<langle>Skip,s'\<rangle> \<and> n = i + j + 2)"
proof(induct c\<equiv>"while (b) cx" s n c'\<equiv>"Skip" s' arbitrary:c rule:red_n.induct)
  case (red_n_Base c s)
  from `c = while (b) cx` `c = Skip` have False by simp
  thus ?case by simp
next
  case (red_n_Rec c s c'' s'' n c' s')
  from `\<langle>c,s\<rangle> \<rightarrow> \<langle>c'',s''\<rangle>` `c = while (b) cx`
  have "(\<lbrakk>b\<rbrakk> s = Some false \<and> c'' = Skip \<and> s'' = s) \<or>
    (\<lbrakk>b\<rbrakk> s = Some true \<and> c'' = cx;;while (b) cx \<and> s'' = s)"
    by(induct rule:red_induct) auto
  thus ?case
  proof
    assume "\<lbrakk>b\<rbrakk> s = Some false \<and> c'' = Skip \<and> s'' = s"
    hence "\<lbrakk>b\<rbrakk> s = Some false" and "c'' = Skip" and "s'' = s" by simp_all
    with `\<langle>c'',s''\<rangle> \<rightarrow>\<^bsup>n\<^esup> \<langle>c',s'\<rangle>` have "s = s'" and "n = 0"
      by(induct rule:red_n.induct,auto elim:red.cases)
    with `\<lbrakk>b\<rbrakk> s = Some false` show ?thesis by fastsimp
  next
    assume "\<lbrakk>b\<rbrakk> s = Some true \<and> c'' = cx;;while (b) cx \<and> s'' = s"
    hence "\<lbrakk>b\<rbrakk> s = Some true" and "c'' = cx;;while (b) cx" 
      and "s'' = s" by simp_all
    with `\<langle>c'',s''\<rangle> \<rightarrow>\<^bsup>n\<^esup> \<langle>c',s'\<rangle>` `c' = Skip` 
    obtain i j sx where "\<langle>cx,s\<rangle> \<rightarrow>\<^bsup>i\<^esup> \<langle>Skip,sx\<rangle>" and "\<langle>while (b) cx,sx\<rangle> \<rightarrow>\<^bsup>j\<^esup> \<langle>Skip,s'\<rangle>"
      and "n = i + j + 1" by(fastsimp elim:Seq_red_nE)
    with `\<lbrakk>b\<rbrakk> s = Some true` show ?thesis by fastsimp
  qed
qed


lemma while_red_n_induct [consumes 1, case_names false true]:
  assumes major: "\<langle>while (b) cx,s\<rangle> \<rightarrow>\<^bsup>n\<^esup> \<langle>Skip,s'\<rangle>"
  and IHfalse:"\<And>s.  \<lbrakk>b\<rbrakk> s = Some false \<Longrightarrow> P s s"
  and IHtrue:"\<And>s i j s''. \<lbrakk>  \<lbrakk>b\<rbrakk> s = Some true; \<langle>cx,s\<rangle> \<rightarrow>\<^bsup>i\<^esup> \<langle>Skip,s''\<rangle>; 
                            \<langle>while (b) cx,s''\<rangle> \<rightarrow>\<^bsup>j\<^esup> \<langle>Skip,s'\<rangle>; P s'' s'\<rbrakk> \<Longrightarrow> P s s'"
  shows "P s s'"
using major
proof(induct n arbitrary:s rule:nat_less_induct)
  fix n s
  assume IHall:"\<forall>m<n. \<forall>x. \<langle>while (b) cx,x\<rangle> \<rightarrow>\<^bsup>m\<^esup> \<langle>Skip,s'\<rangle> \<longrightarrow> P x s'"
    and  "\<langle>while (b) cx,s\<rangle> \<rightarrow>\<^bsup>n\<^esup> \<langle>Skip,s'\<rangle>"
  from `\<langle>while (b) cx,s\<rangle> \<rightarrow>\<^bsup>n\<^esup> \<langle>Skip,s'\<rangle>`
  have "(\<lbrakk>b\<rbrakk> s = Some false \<and> s = s' \<and> n = 1) \<or>
     (\<exists>i j s''. \<lbrakk>b\<rbrakk> s = Some true \<and> \<langle>cx,s\<rangle> \<rightarrow>\<^bsup>i\<^esup> \<langle>Skip,s''\<rangle> \<and> 
                \<langle>while (b) cx,s''\<rangle> \<rightarrow>\<^bsup>j\<^esup> \<langle>Skip,s'\<rangle> \<and> n = i + j + 2)" 
    by(rule while_red_nE)
  thus "P s s'"
  proof
    assume "\<lbrakk>b\<rbrakk> s = Some false \<and> s = s' \<and> n = 1"
    hence "\<lbrakk>b\<rbrakk> s = Some false" and "s = s'" by auto
    from IHfalse[OF `\<lbrakk>b\<rbrakk> s = Some false`] have "P s s" .
    with `s = s'` show ?thesis by simp
  next
    assume "\<exists>i j s''. \<lbrakk>b\<rbrakk> s = Some true \<and> \<langle>cx,s\<rangle> \<rightarrow>\<^bsup>i\<^esup> \<langle>Skip,s''\<rangle> \<and> 
                      \<langle>while (b) cx,s''\<rangle> \<rightarrow>\<^bsup>j\<^esup> \<langle>Skip,s'\<rangle> \<and> n = i + j + 2"
    then obtain i j s'' where "\<lbrakk>b\<rbrakk> s = Some true"
      and "\<langle>cx,s\<rangle> \<rightarrow>\<^bsup>i\<^esup> \<langle>Skip,s''\<rangle>" and "\<langle>while (b) cx,s''\<rangle> \<rightarrow>\<^bsup>j\<^esup> \<langle>Skip,s'\<rangle>"
      and "n = i + j + 2" by blast
    with IHall have "P s'' s'"
      apply(erule_tac x="j" in allE) apply clarsimp done
    from IHtrue[OF `\<lbrakk>b\<rbrakk> s = Some true` `\<langle>cx,s\<rangle> \<rightarrow>\<^bsup>i\<^esup> \<langle>Skip,s''\<rangle>` 
      `\<langle>while (b) cx,s''\<rangle> \<rightarrow>\<^bsup>j\<^esup> \<langle>Skip,s'\<rangle>` this] show ?thesis .
  qed
qed


lemma reds_to_red_n:"\<langle>c,s\<rangle> \<rightarrow>* \<langle>c',s'\<rangle> \<Longrightarrow> \<exists>n. \<langle>c,s\<rangle> \<rightarrow>\<^bsup>n\<^esup> \<langle>c',s'\<rangle>"
by(induct rule:converse_rtranclp_induct2,auto intro:red_n.intros)


lemma red_n_to_reds:"\<langle>c,s\<rangle> \<rightarrow>\<^bsup>n\<^esup> \<langle>c',s'\<rangle> \<Longrightarrow> \<langle>c,s\<rangle> \<rightarrow>* \<langle>c',s'\<rangle>"
by(induct rule:red_n.induct,auto intro:converse_rtranclp_into_rtranclp)


lemma while_reds_induct[consumes 1, case_names false true]:
  "\<lbrakk>\<langle>while (b) cx,s\<rangle> \<rightarrow>* \<langle>Skip,s'\<rangle>; \<And>s. \<lbrakk>b\<rbrakk> s = Some false \<Longrightarrow> P s s;
    \<And>s s''. \<lbrakk> \<lbrakk>b\<rbrakk> s = Some true; \<langle>cx,s\<rangle> \<rightarrow>* \<langle>Skip,s''\<rangle>; 
             \<langle>while (b) cx,s''\<rangle> \<rightarrow>* \<langle>Skip,s'\<rangle>; P s'' s'\<rbrakk> \<Longrightarrow> P s s'\<rbrakk>
  \<Longrightarrow> P s s'"
apply(drule reds_to_red_n,clarsimp)
apply(erule while_red_n_induct,clarsimp)
by(auto dest:red_n_to_reds)


lemma red_det:
  "\<lbrakk>\<langle>c,s\<rangle> \<rightarrow> \<langle>c\<^isub>1,s\<^isub>1\<rangle>; \<langle>c,s\<rangle> \<rightarrow> \<langle>c\<^isub>2,s\<^isub>2\<rangle>\<rbrakk> \<Longrightarrow> c\<^isub>1 = c\<^isub>2 \<and> s\<^isub>1 = s\<^isub>2"
proof(induct arbitrary:c\<^isub>2 rule:red_induct)
  case (SeqRed c\<^isub>1 s c\<^isub>1' s' c\<^isub>2')
  note IH = `\<And>c\<^isub>2. \<langle>c\<^isub>1,s\<rangle> \<rightarrow> \<langle>c\<^isub>2,s\<^isub>2\<rangle> \<Longrightarrow> c\<^isub>1' = c\<^isub>2 \<and> s' = s\<^isub>2`
  from `\<langle>c\<^isub>1;;c\<^isub>2',s\<rangle> \<rightarrow> \<langle>c\<^isub>2,s\<^isub>2\<rangle>` have "c\<^isub>1 = Skip \<or> (\<exists>cx. c\<^isub>2 = cx;;c\<^isub>2' \<and> \<langle>c\<^isub>1,s\<rangle> \<rightarrow> \<langle>cx,s\<^isub>2\<rangle>)"
    by(fastsimp elim:red.cases)
  thus ?case
  proof
    assume "c\<^isub>1 = Skip"
    with `\<langle>c\<^isub>1,s\<rangle> \<rightarrow> \<langle>c\<^isub>1',s'\<rangle>` have False by(fastsimp elim:red.cases)
    thus ?thesis by simp
  next
    assume "\<exists>cx. c\<^isub>2 = cx;;c\<^isub>2' \<and> \<langle>c\<^isub>1,s\<rangle> \<rightarrow> \<langle>cx,s\<^isub>2\<rangle>"
    then obtain cx where "c\<^isub>2 = cx;;c\<^isub>2'" and "\<langle>c\<^isub>1,s\<rangle> \<rightarrow> \<langle>cx,s\<^isub>2\<rangle>" by blast
    from IH[OF `\<langle>c\<^isub>1,s\<rangle> \<rightarrow> \<langle>cx,s\<^isub>2\<rangle>`] have "c\<^isub>1' = cx \<and> s' = s\<^isub>2" .
    with `c\<^isub>2 = cx;;c\<^isub>2'` show ?thesis by simp
  qed
qed (fastsimp elim:red.cases)+


theorem reds_det:
  "\<lbrakk>\<langle>c,s\<rangle> \<rightarrow>* \<langle>Skip,s\<^isub>1\<rangle>; \<langle>c,s\<rangle> \<rightarrow>* \<langle>Skip,s\<^isub>2\<rangle>\<rbrakk> \<Longrightarrow> s\<^isub>1 = s\<^isub>2"
proof(induct rule:converse_rtranclp_induct2)
  case refl
  from `\<langle>Skip,s\<^isub>1\<rangle> \<rightarrow>* \<langle>Skip,s\<^isub>2\<rangle>` show ?case
    by -(erule converse_rtranclpE,auto elim:red.cases)
next
  case (step c'' s'' c' s')
  note IH = `\<langle>c',s'\<rangle> \<rightarrow>* \<langle>Skip,s\<^isub>2\<rangle> \<Longrightarrow> s\<^isub>1 = s\<^isub>2`
  from step have "\<langle>c'',s''\<rangle> \<rightarrow> \<langle>c',s'\<rangle>"
    by simp
  from `\<langle>c'',s''\<rangle> \<rightarrow>* \<langle>Skip,s\<^isub>2\<rangle>` this have "\<langle>c',s'\<rangle> \<rightarrow>* \<langle>Skip,s\<^isub>2\<rangle>"
    by -(erule converse_rtranclpE,auto elim:red.cases dest:red_det)
  from IH[OF this] show ?thesis .
qed


end
